const char buffer_lua[] =
"-- buffer.lua (internal file)\n"
"\n"
"local ffi = require('ffi')\n"
"local READAHEAD = 16320\n"
"\n"
"ffi.cdef[[\n"
"struct slab_cache;\n"
"struct slab_cache *\n"
"tarantool_lua_slab_cache();\n"
"\n"
"struct ibuf *\n"
"cord_ibuf_take(void);\n"
"\n"
"void\n"
"cord_ibuf_put(struct ibuf *ibuf);\n"
"\n"
"void\n"
"cord_ibuf_drop(struct ibuf *ibuf);\n"
"\n"
"struct ibuf\n"
"{\n"
"    struct slab_cache *slabc;\n"
"    char *buf;\n"
"    /** Start of input. */\n"
"    char *rpos;\n"
"    /** End of useful input */\n"
"    char *wpos;\n"
"    /** End of ibuf. */\n"
"    char *epos;\n"
"    size_t start_capacity;\n"
"};\n"
"\n"
"void\n"
"ibuf_create(struct ibuf *ibuf, struct slab_cache *slabc, size_t start_capacity);\n"
"\n"
"void\n"
"ibuf_destroy(struct ibuf *ibuf);\n"
"\n"
"void\n"
"ibuf_reinit(struct ibuf *ibuf);\n"
"\n"
"void *\n"
"ibuf_reserve_slow(struct ibuf *ibuf, size_t size);\n"
"]]\n"
"\n"
"local builtin = ffi.C\n"
"local ibuf_t = ffi.typeof('struct ibuf')\n"
"\n"
"local function errorf(s, ...)\n"
"    error(string.format(s, ...))\n"
"end\n"
"\n"
"local function checkibuf(buf, method)\n"
"    if not ffi.istype(ibuf_t, buf) then\n"
"        errorf('Attempt to call method without object, use ibuf:%s()', method)\n"
"    end\n"
"end\n"
"\n"
"local function ibuf_capacity(buf)\n"
"    checkibuf(buf, 'capacity')\n"
"    return tonumber(buf.epos - buf.buf)\n"
"end\n"
"\n"
"local function ibuf_pos(buf)\n"
"    checkibuf(buf, 'pos')\n"
"    return tonumber(buf.rpos - buf.buf)\n"
"end\n"
"\n"
"local function ibuf_used(buf)\n"
"    checkibuf(buf, 'size')\n"
"    return tonumber(buf.wpos - buf.rpos)\n"
"end\n"
"\n"
"local function ibuf_unused(buf)\n"
"    checkibuf(buf, 'unused')\n"
"    return tonumber(buf.epos - buf.wpos)\n"
"end\n"
"\n"
"local function ibuf_recycle(buf)\n"
"    checkibuf(buf, 'recycle')\n"
"    builtin.ibuf_reinit(buf)\n"
"end\n"
"\n"
"local function ibuf_reset(buf)\n"
"    checkibuf(buf, 'reset')\n"
"    buf.rpos = buf.buf\n"
"    buf.wpos = buf.buf\n"
"end\n"
"\n"
"local function ibuf_reserve_slow(buf, size)\n"
"    local ptr = builtin.ibuf_reserve_slow(buf, size)\n"
"    if ptr == nil then\n"
"        errorf(\"Failed to allocate %d bytes in ibuf\", size)\n"
"    end\n"
"    return ffi.cast('char *', ptr)\n"
"end\n"
"\n"
"local function ibuf_reserve(buf, size)\n"
"    checkibuf(buf, 'reserve')\n"
"    if buf.wpos + size <= buf.epos then\n"
"        return buf.wpos\n"
"    end\n"
"    return ibuf_reserve_slow(buf, size)\n"
"end\n"
"\n"
"local function ibuf_alloc(buf, size)\n"
"    checkibuf(buf, 'alloc')\n"
"    local wpos\n"
"    if buf.wpos + size <= buf.epos then\n"
"        wpos = buf.wpos\n"
"    else\n"
"        wpos = ibuf_reserve_slow(buf, size)\n"
"    end\n"
"    buf.wpos = buf.wpos + size\n"
"    return wpos\n"
"end\n"
"\n"
"local function checksize(buf, size)\n"
"    if buf.rpos + size > buf.wpos then\n"
"        errorf(\"Attempt to read out of range bytes: needed=%d size=%d\",\n"
"            tonumber(size), ibuf_used(buf))\n"
"    end\n"
"end\n"
"\n"
"local function ibuf_checksize(buf, size)\n"
"    checkibuf(buf, 'checksize')\n"
"    checksize(buf, size)\n"
"    return buf.rpos\n"
"end\n"
"\n"
"local function ibuf_read(buf, size)\n"
"    checkibuf(buf, 'read')\n"
"    checksize(buf, size)\n"
"    local rpos = buf.rpos\n"
"    buf.rpos = rpos + size\n"
"    return rpos\n"
"end\n"
"\n"
"local function ibuf_serialize(buf)\n"
"    local properties = { rpos = buf.rpos, wpos = buf.wpos }\n"
"    return { ibuf = properties }\n"
"end\n"
"\n"
"local ibuf_methods = {\n"
"    recycle = ibuf_recycle;\n"
"    reset = ibuf_reset;\n"
"\n"
"    reserve = ibuf_reserve;\n"
"    alloc = ibuf_alloc;\n"
"\n"
"    checksize = ibuf_checksize;\n"
"    read = ibuf_read;\n"
"    __serialize = ibuf_serialize;\n"
"\n"
"    size = ibuf_used;\n"
"    capacity = ibuf_capacity;\n"
"    pos = ibuf_pos;\n"
"    unused = ibuf_unused;\n"
"}\n"
"\n"
"local function ibuf_tostring(self)\n"
"    return '<ibuf>'\n"
"end\n"
"local ibuf_mt = {\n"
"    __gc = ibuf_recycle;\n"
"    __index = ibuf_methods;\n"
"    __tostring = ibuf_tostring;\n"
"};\n"
"\n"
"ffi.metatype(ibuf_t, ibuf_mt);\n"
"\n"
"local function ibuf_new(arg)\n"
"    local buf = ffi.new(ibuf_t)\n"
"    local slabc = builtin.tarantool_lua_slab_cache()\n"
"    builtin.ibuf_create(buf, slabc, READAHEAD)\n"
"    if arg == nil then\n"
"        return buf\n"
"    elseif type(arg) == 'number' then\n"
"        ibuf_reserve(buf, arg)\n"
"        return buf\n"
"    end\n"
"    errorf('Usage: ibuf([size])')\n"
"end\n"
"\n"
"--\n"
"-- Stash keeps an FFI object for re-usage and helps to ensure the proper\n"
"-- ownership. Is supposed to be used in yield-free code when almost always it is\n"
"-- possible to put the taken object back.\n"
"-- Then cost of the stash is almost the same as ffi.new() for small objects like\n"
"-- 'int[1]' even when jitted. Examples:\n"
"--\n"
"-- * ffi.new('int[1]') is about ~0.4ns, while the stash take() + put() is about\n"
"--   ~0.8ns;\n"
"--\n"
"-- * Much better on objects > 128 bytes in size. ffi.new('struct uri[1]') is\n"
"--   ~300ns, while the stash is still ~0.8ns;\n"
"--\n"
"-- * For structs not allocated as an array is also much better than ffi.new().\n"
"--   For instance, ffi.new('struct tt_uuid') is ~300ns, the stash is ~0.8ns.\n"
"--   Even though 'struct tt_uuid' is 16 bytes;\n"
"--\n"
"local function ffi_stash_new(c_type)\n"
"    local item = nil\n"
"\n"
"    local function take()\n"
"        local res\n"
"        -- This line is guaranteed to be GC-safe. GC is not invoked. Because\n"
"        -- there are no allocation. So it can be considered 'atomic'.\n"
"        res, item = item, nil\n"
"        -- The next lines don't need to be atomic and can survive GC. The only\n"
"        -- important part was to take the global item and set it to nil.\n"
"        if res then\n"
"            return res\n"
"        end\n"
"        return ffi.new(c_type)\n"
"    end\n"
"\n"
"    local function put(i)\n"
"        -- It is ok to rewrite the existing global item if it was set. Does\n"
"        -- not matter. They are all the same.\n"
"        item = i\n"
"    end\n"
"\n"
"    -- Due to some random reason if the stash returns a table with methods it\n"
"    -- works faster than returning them as multiple values. Regardless of how\n"
"    -- the methods are used later. Even if the caller will cache take and put\n"
"    -- methods anyway.\n"
"    return {\n"
"        take = take,\n"
"        put = put,\n"
"    }\n"
"end\n"
"\n"
"--\n"
"-- Cord buffer is useful for the places, where\n"
"--\n"
"-- * Want to reuse the already allocated memory which might be stored in the\n"
"--   cord buf. Although sometimes the buffer is recycled, so should not rely on\n"
"--   being able to reuse it always. When reused, the win is the biggest -\n"
"--   becomes about x20 times faster than a new buffer creation (~5ns vs ~100ns);\n"
"--\n"
"-- * Want to avoid allocation of a new ibuf because it produces a new GC object\n"
"--   which is additional load for Lua GC. Although according to benches it is\n"
"--   not super expensive;\n"
"--\n"
"-- * Almost always can put the buffer back manually. Not rely on it being\n"
"--   recycled automatically. It is recycled, but still should not rely on that;\n"
"--\n"
"-- It is important to wrap the C functions, not expose them directly. Because\n"
"-- JIT works a bit better when C functions are called as 'ffi.C.func()' than\n"
"-- 'func()' with func being cached. The only pros is to cache 'ffi.C' itself.\n"
"-- It is quite strange though how having them wrapped into a Lua function is\n"
"-- faster than cached directly as C functions.\n"
"--\n"
"local function cord_ibuf_take()\n"
"    return builtin.cord_ibuf_take()\n"
"end\n"
"\n"
"local function cord_ibuf_put(buf)\n"
"    return builtin.cord_ibuf_put(buf)\n"
"end\n"
"\n"
"local function cord_ibuf_drop(buf)\n"
"    return builtin.cord_ibuf_drop(buf)\n"
"end\n"
"\n"
"local internal = {\n"
"    cord_ibuf_take = cord_ibuf_take,\n"
"    cord_ibuf_put = cord_ibuf_put,\n"
"    cord_ibuf_drop = cord_ibuf_drop,\n"
"}\n"
"\n"
"return {\n"
"    internal = internal,\n"
"    ibuf = ibuf_new;\n"
"    READAHEAD = READAHEAD;\n"
"    ffi_stash_new = ffi_stash_new,\n"
"}\n"
""
;
