const char digest_lua[] =
"-- digest.lua (internal file)\n"
"\n"
"local ffi = require('ffi')\n"
"local crypto = require('crypto')\n"
"local bit = require('bit')\n"
"local buffer = require('buffer')\n"
"local cord_ibuf_take = buffer.internal.cord_ibuf_take\n"
"local cord_ibuf_put = buffer.internal.cord_ibuf_put\n"
"\n"
"ffi.cdef[[\n"
"    /* internal implementation */\n"
"    unsigned char *SHA1internal(const unsigned char *d, size_t n, unsigned char *md);\n"
"\n"
"    /* from libc */\n"
"    int snprintf(char *str, size_t size, const char *format, ...);\n"
"\n"
"    typedef uint32_t (*crc32_func)(uint32_t crc,\n"
"        const unsigned char *buf, unsigned int len);\n"
"    extern int32_t guava(uint64_t state, int32_t buckets);\n"
"    extern crc32_func crc32_calc;\n"
"\n"
"    /* base64 */\n"
"    int base64_bufsize(int binsize, int options);\n"
"    int base64_decode(const char *in_base64, int in_len, char *out_bin, int out_len);\n"
"    int base64_encode(const char *in_bin, int in_len, char *out_base64, int out_len, int options);\n"
"\n"
"    /* random */\n"
"    void random_bytes(char *, size_t);\n"
"\n"
"    /* from third_party/PMurHash.h */\n"
"    void PMurHash32_Process(uint32_t *ph1, uint32_t *pcarry, const void *key, int len);\n"
"    uint32_t PMurHash32_Result(uint32_t h1, uint32_t carry, uint32_t total_length);\n"
"    uint32_t PMurHash32(uint32_t seed, const void *key, int len);\n"
"]]\n"
"\n"
"-- @sa base64.h\n"
"local BASE64_NOPAD = 1\n"
"local BASE64_NOWRAP = 2\n"
"local BASE64_URLSAFE = 7\n"
"\n"
"local digest_shortcuts = {\n"
"    sha224  = 'SHA224',\n"
"    sha256  = 'SHA256',\n"
"    sha384  = 'SHA384',\n"
"    sha512  = 'SHA512',\n"
"    md5     = 'MD5',\n"
"    md4     = 'MD4',\n"
"}\n"
"local internal = require(\"digest\")\n"
"\n"
"local PMurHash\n"
"local PMurHash_methods = {\n"
"\n"
"    update = function(self, str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage: murhash:update(string)\")\n"
"        end\n"
"        ffi.C.PMurHash32_Process(self.seed, self.value, str, string.len(str))\n"
"        self.total_length = self.total_length + string.len(str)\n"
"    end,\n"
"\n"
"    result = function(self)\n"
"        return ffi.C.PMurHash32_Result(self.seed[0], self.value[0], self.total_length)\n"
"    end,\n"
"\n"
"    clear = function(self)\n"
"        self.seed[0] = self.default_seed\n"
"        self.total_length = 0\n"
"        self.value[0] = 0\n"
"    end,\n"
"\n"
"    copy = function(self)\n"
"        local new_self = PMurHash.new()\n"
"        new_self.seed[0] = self.seed[0]\n"
"        new_self.value[0] = self.value[0]\n"
"        new_self.total_length = self.total_length\n"
"        return new_self\n"
"    end\n"
"}\n"
"\n"
"PMurHash = {\n"
"    default_seed = 13,\n"
"\n"
"    new = function(opts)\n"
"        opts = opts or {}\n"
"        local self = setmetatable({}, { __index = PMurHash_methods })\n"
"        self.default_seed = (opts.seed or PMurHash.default_seed)\n"
"        self.seed = ffi.new(\"int[1]\", self.default_seed)\n"
"        self.value = ffi.new(\"int[1]\", 0)\n"
"        self.total_length = 0\n"
"        return self\n"
"    end\n"
"}\n"
"\n"
"setmetatable(PMurHash, {\n"
"    __call = function(self, str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage: digest.murhash(string)\")\n"
"        end\n"
"        return ffi.C.PMurHash32(PMurHash.default_seed, str, string.len(str))\n"
"    end\n"
"})\n"
"\n"
"local CRC32\n"
"local CRC32_methods = {\n"
"    update = function(self, str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage crc32:update(string)\")\n"
"        end\n"
"        self.value = ffi.C.crc32_calc(self.value, str, string.len(str))\n"
"    end,\n"
"\n"
"    result = function(self)\n"
"        return self.value\n"
"    end,\n"
"\n"
"    clear = function(self)\n"
"        self.value = CRC32.crc_begin\n"
"    end,\n"
"\n"
"    copy = function(self)\n"
"        local new_self = CRC32.new()\n"
"        new_self.value = self.value\n"
"        return new_self\n"
"    end\n"
"}\n"
"\n"
"CRC32 = {\n"
"    crc_begin = 4294967295,\n"
"\n"
"    new = function()\n"
"        local self = setmetatable({}, { __index = CRC32_methods })\n"
"        self.value = CRC32.crc_begin\n"
"        return self\n"
"    end\n"
"}\n"
"\n"
"setmetatable(CRC32, {\n"
"    __call = function(self, str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage digest.crc32(string)\")\n"
"        end\n"
"        return ffi.C.crc32_calc(CRC32.crc_begin, str, string.len(str))\n"
"    end\n"
"})\n"
"\n"
"local pbkdf2 = function(pass, salt, iters, digest_len)\n"
"    if type(pass) ~= 'string' or type(salt) ~= 'string' then\n"
"        error(\"Usage: digest.pbkdf2(pass, salt[,iters][,digest_len])\")\n"
"    end\n"
"    if iters and type(iters) ~= 'number' then\n"
"        error(\"iters must be a number\")\n"
"    end\n"
"    if digest_len and type(digest_len) ~= 'number' then\n"
"        error(\"digest_len must be a number\")\n"
"    end\n"
"    iters = iters or 100000\n"
"    digest_len = digest_len or 128\n"
"    if digest_len > 128 then\n"
"        error(\"too big digest size\")\n"
"    end\n"
"    return internal.pbkdf2(pass, salt, iters, digest_len)\n"
"end\n"
"\n"
"local m = {\n"
"    base64_encode = function(bin, options)\n"
"        if type(bin) ~= 'string' or\n"
"           options ~= nil and type(options) ~= 'table' then\n"
"            error('Usage: digest.base64_encode(string[, table])')\n"
"        end\n"
"        local mask = 0\n"
"        if options ~= nil then\n"
"            if options.urlsafe then\n"
"                mask = bit.bor(mask, BASE64_URLSAFE)\n"
"            end\n"
"            if options.nopad then\n"
"                mask = bit.bor(mask, BASE64_NOPAD)\n"
"            end\n"
"            if options.nowrap then\n"
"                mask = bit.bor(mask, BASE64_NOWRAP)\n"
"            end\n"
"        end\n"
"        local blen = #bin\n"
"        local slen = ffi.C.base64_bufsize(blen, mask)\n"
"        local ibuf = cord_ibuf_take()\n"
"        local str = ibuf:alloc(slen)\n"
"        local len = ffi.C.base64_encode(bin, blen, str, slen, mask)\n"
"        str = ffi.string(str, len)\n"
"        cord_ibuf_put(ibuf)\n"
"        return str\n"
"    end,\n"
"\n"
"    base64_decode = function(str)\n"
"        if type(str) ~= 'string' then\n"
"            error('Usage: digest.base64_decode(string)')\n"
"        end\n"
"        local slen = #str\n"
"        local blen = math.ceil(slen * 3 / 4)\n"
"        local ibuf = cord_ibuf_take()\n"
"        local bin = ibuf:alloc(blen)\n"
"        local len = ffi.C.base64_decode(str, slen, bin, blen)\n"
"        bin = ffi.string(bin, len)\n"
"        cord_ibuf_put(ibuf)\n"
"        return bin\n"
"    end,\n"
"\n"
"    crc32 = CRC32,\n"
"\n"
"    crc32_update = function(crc, str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage: digest.crc32_update(string)\")\n"
"        end\n"
"        return ffi.C.crc32_calc(tonumber(crc), str, string.len(str))\n"
"    end,\n"
"\n"
"    sha1 = function(str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage: digest.sha1(string)\")\n"
"        end\n"
"        local r = ffi.C.SHA1internal(str, #str, nil)\n"
"        return ffi.string(r, 20)\n"
"    end,\n"
"\n"
"    sha1_hex = function(str)\n"
"        if type(str) ~= 'string' then\n"
"            error(\"Usage: digest.sha1_hex(string)\")\n"
"        end\n"
"        local r = ffi.C.SHA1internal(str, #str, nil)\n"
"        return string.hex(ffi.string(r, 20))\n"
"    end,\n"
"\n"
"    guava = function(state, buckets)\n"
"       return ffi.C.guava(state, buckets)\n"
"    end,\n"
"\n"
"    urandom = function(n)\n"
"        if n == nil then\n"
"            error('Usage: digest.urandom(len)')\n"
"        end\n"
"        local ibuf = cord_ibuf_take()\n"
"        local buf = ibuf:alloc(n)\n"
"        ffi.C.random_bytes(buf, n)\n"
"        buf = ffi.string(buf, n)\n"
"        cord_ibuf_put(ibuf)\n"
"        return buf\n"
"    end,\n"
"\n"
"    murmur = PMurHash,\n"
"\n"
"    pbkdf2 = pbkdf2,\n"
"\n"
"    pbkdf2_hex = function(pass, salt, iters, digest_len)\n"
"        if type(pass) ~= 'string' or type(salt) ~= 'string' then\n"
"            error(\"Usage: digest.pbkdf2_hex(pass, salt)\")\n"
"        end\n"
"        return string.hex(pbkdf2(pass, salt, iters, digest_len))\n"
"    end\n"
"}\n"
"\n"
"for digest, _ in pairs(digest_shortcuts) do\n"
"    m[digest] = function (str)\n"
"        return crypto.digest[digest](str)\n"
"    end\n"
"    m[digest .. '_hex'] = function (str)\n"
"        if type(str) ~= 'string' then\n"
"            error('Usage: digest.'..digest..'_hex(string)')\n"
"        end\n"
"        return string.hex(crypto.digest[digest](str))\n"
"    end\n"
"end\n"
"\n"
"m['aes256cbc'] = {\n"
"    encrypt = function (str, key, iv)\n"
"        return crypto.cipher.aes256.cbc.encrypt(str, key, iv)\n"
"    end,\n"
"    decrypt = function (str, key, iv)\n"
"        return crypto.cipher.aes256.cbc.decrypt(str, key, iv)\n"
"    end\n"
"}\n"
"\n"
"return m\n"
""
;
