const char httpc_lua[] =
"--\n"
"--  Copyright (C) 2016-2017 Tarantool AUTHORS: please see AUTHORS file.\n"
"--\n"
"--  Redistribution and use in source and binary forms, with or\n"
"--  without modification, are permitted provided that the following\n"
"--  conditions are met:\n"
"--\n"
"--  1. Redistributions of source code must retain the above\n"
"--   copyright notice, this list of conditions and the\n"
"--   following disclaimer.\n"
"--\n"
"--  2. Redistributions in binary form must reproduce the above\n"
"--   copyright notice, this list of conditions and the following\n"
"--   disclaimer in the documentation and/or other materials\n"
"--   provided with the distribution.\n"
"--\n"
"--  THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND\n"
"--  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED\n"
"--  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR\n"
"--  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL\n"
"--  <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,\n"
"--  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL\n"
"--  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF\n"
"--  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR\n"
"--  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF\n"
"--  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n"
"--  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF\n"
"--  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF\n"
"--  SUCH DAMAGE.\n"
"--\n"
"\n"
"local driver = package.loaded.http.client\n"
"package.loaded.http = nil\n"
"\n"
"local curl_mt\n"
"\n"
"--\n"
"--  <http> - create a new curl instance.\n"
"--\n"
"--  Parameters:\n"
"--\n"
"--  max_connections -  Maximum number of entries in the connection cache\n"
"--  max_total_connections -  Maximum number of active connections\n"
"--\n"
"--  Returns:\n"
"--  curl object or raise error()\n"
"--\n"
"\n"
"local http_new = function(opts)\n"
"\n"
"    opts = opts or {}\n"
"\n"
"    opts.max_connections = opts.max_connections or -1\n"
"    opts.max_total_connections = opts.max_total_connections or 0\n"
"\n"
"    local curl = driver.new(opts.max_connections, opts.max_total_connections)\n"
"    return setmetatable({ curl = curl, }, curl_mt )\n"
"end\n"
"\n"
"local check_args_fmt = 'Use client:%s(...) instead of client.%s(...):'\n"
"\n"
"local function check_args(self, method)\n"
"    if type(self) ~= 'table' then\n"
"        error(check_args_fmt:format(method, method), 2)\n"
"    end\n"
"end\n"
"\n"
"--\n"
"-- RFC2616: http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2\n"
"--\n"
"-- Multiple message-header fields with the same field-name MAY be present\n"
"-- in a message if and only if the entire field-value for that header field\n"
"-- is defined as a comma-separated list [i.e., #(values)]. It MUST be possible\n"
"-- to combine the multiple header fields into one \"field-name: field-value\"\n"
"-- pair, without changing the semantics of the message, by appending each\n"
"-- subsequent field-value to the first, each separated by a comma. The order\n"
"-- in which header fields with the same field-name are received is therefore\n"
"-- significant to the interpretation of the combined field value, and thus\n"
"-- a proxy MUST NOT change the order of these field values when a message\n"
"-- is forwarded.\n"
"--\n"
"-- Tarantool implementation concatenates all headers by default except\n"
"-- the blacklist below.\n"
"--\n"
"\n"
"local special_headers = {\n"
"    [\"age\"] = true,\n"
"    [\"authorization\"] = true,\n"
"    [\"content-length\"] = true,\n"
"    [\"content-type\"] = true,\n"
"    [\"etag\"] = true,\n"
"    [\"expires\"] = true,\n"
"    [\"from\"] = true,\n"
"    [\"host\"] = true,\n"
"    [\"if-modified-since\"] = true,\n"
"    [\"if-unmodified-since\"] = true,\n"
"    [\"last-modified\"] = true,\n"
"    [\"location\"] = true,\n"
"    [\"max-forwards\"] = true,\n"
"    [\"proxy-authorization\"] = true,\n"
"    [\"referer\"] = true,\n"
"    [\"retry-after\"] = true,\n"
"    [\"user-agent\"] = true,\n"
"}\n"
"\n"
"local special_characters = {\n"
"    ['('] = true,\n"
"    [')'] = true,\n"
"    ['['] = true,\n"
"    [']'] = true,\n"
"    ['<'] = true,\n"
"    ['>'] = true,\n"
"    ['@'] = true,\n"
"    [','] = true,\n"
"    [';'] = true,\n"
"    [':'] = true,\n"
"    ['\\\\'] = true,\n"
"    ['\\\"'] = true,\n"
"    ['/'] = true,\n"
"    ['\?'] = true,\n"
"    ['='] = true,\n"
"    ['{'] = true,\n"
"    ['}'] = true,\n"
"    [' '] = true,\n"
"    ['\\t'] = true\n"
"}\n"
"\n"
"local option_keys = {\n"
"    [\"Expires\"] = true,\n"
"    [\"Max-Age\"] = true,\n"
"    [\"Domain\"] = true,\n"
"    [\"Path\"] = true,\n"
"    [\"Secure\"] = true,\n"
"    [\"HttpOnly\"] = true,\n"
"    [\"SameSite\"] = true,\n"
"}\n"
"\n"
"--local function process_set_cookies(value, result)\n"
"--    local key_start, value_start\n"
"--    local key, val\n"
"--    local symbols = value:gmatch('.')\n"
"--    local options = {}\n"
"--    local cur = 0\n"
"--    for v in symbols do\n"
"--        cur = cur + 1\n"
"--        if v == ' ' or v == '\\t' then\n"
"--            goto continue\n"
"--        end\n"
"--        key_start = cur\n"
"--        -- parse cookie name\n"
"--        while not special_characters[v] do\n"
"--            if v == nil then\n"
"--                return\n"
"--            end\n"
"--            v = symbols()\n"
"--            cur = cur + 1\n"
"--        end\n"
"--        key = value:sub(key_start, cur)\n"
"--        if not v or v ~= '=' then\n"
"--            -- invalid header\n"
"--            return\n"
"--        end\n"
"--        while v == ' ' do\n"
"--            v = symbols()\n"
"--            cur = cur + 1\n"
"--        end\n"
"--\n"
"--        if v == nil then\n"
"--            return\n"
"--        end\n"
"--\n"
"--        while v and v ~= ';' do\n"
"--            if v == nil then\n"
"--                break\n"
"--            end\n"
"--            v = symbols()\n"
"--            cur = cur + 1\n"
"--        end\n"
"--\n"
"--        result[key] = {val, options}\n"
"--        ::continue::\n"
"--    end\n"
"--end\n"
"\n"
"local function process_cookie(cookie, result)\n"
"    local vals = cookie:split(';')\n"
"    local val = vals[1]:split('=')\n"
"    if #val < 2 then\n"
"        return\n"
"    end\n"
"    val[1] = string.strip(val[1])\n"
"    for c in val[1]:gmatch('.') do\n"
"        if special_characters[c] then\n"
"            return\n"
"        end\n"
"    end\n"
"\n"
"    local options = {}\n"
"    table.remove(vals, 1)\n"
"    for _, opt in pairs(vals) do\n"
"        local splitted = opt:split('=')\n"
"        splitted = string.strip(splitted[1])\n"
"        if option_keys[splitted] then\n"
"            table.insert(options, string.strip(opt))\n"
"        end\n"
"    end\n"
"    result[val[1]] = {string.strip(val[2]), options}\n"
"end\n"
"\n"
"local function process_cookies(cookies)\n"
"    local result = {}\n"
"    for _, val in pairs(cookies) do\n"
"        process_cookie(val, result)\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function process_headers(headers)\n"
"    for header, value in pairs(headers) do\n"
"        if type(value) == 'table' then\n"
"            if special_headers[header] then\n"
"                headers[header] = value[1]\n"
"            else\n"
"                headers[header] = table.concat(value, ',')\n"
"            end\n"
"        end\n"
"    end\n"
"    return headers\n"
"end\n"
"\n"
"--\n"
"--  <request> This function does HTTP request\n"
"--\n"
"--  Parameters:\n"
"--\n"
"--  method  - HTTP method, like GET, POST, PUT and so on\n"
"--  url     - HTTP url, like https://tarantool.org/doc\n"
"--  body    - this parameter is optional, you may use it for passing\n"
"--            data to a server. Like 'My text string!'\n"
"--  options - this is a table of options.\n"
"--\n"
"--      ca_path - a path to ssl certificate dir;\n"
"--\n"
"--      ca_file - a path to ssl certificate file;\n"
"--\n"
"--      unix_socket - a path to Unix domain socket;\n"
"--\n"
"--      verify_host - set on/off verification of the certificate's name (CN)\n"
"--          against host;\n"
"--\n"
"--      verify_peer - set on/off verification of the peer's SSL certificate;\n"
"--\n"
"--      ssl_key - set path to the file with private key for TLS and SSL client\n"
"--          certificate;\n"
"--\n"
"--      ssl_cert - set path to the file with SSL client certificate;\n"
"--\n"
"--      proxy - set a proxy to use;\n"
"--\n"
"--      proxy_port - set a port number the proxy listens on;\n"
"--\n"
"--      proxy_user_pwd - set a user name and a password to use\n"
"--          in authentication;\n"
"--\n"
"--      no_proxy - disable proxy use for specific hosts;\n"
"--\n"
"--      headers - a table of HTTP headers;\n"
"--\n"
"--      keepalive_idle & keepalive_interval -\n"
"--          non-universal keepalive knobs (Linux, AIX, HP-UX, more);\n"
"--\n"
"--      low_speed_time & low_speed_limit -\n"
"--          If the download receives less than\n"
"--          \"low speed limit\" bytes/second\n"
"--          during \"low speed time\" seconds,\n"
"--          the operations is aborted.\n"
"--          You could i.e if you have\n"
"--          a pretty high speed connection, abort if\n"
"--          it is less than 2000 bytes/sec\n"
"--          during 20 seconds;\n"
"--\n"
"--      timeout - time-out the read operation and\n"
"--          waiting for the curl api request\n"
"--          after this amount of seconds;\n"
"--\n"
"--      max_header_name_length - maximum length of a response header;\n"
"--\n"
"--      verbose - set on/off verbose mode;\n"
"--\n"
"--      interface - source interface for outgoing traffic;\n"
"--\n"
"--      follow_location - whether the client will follow\n"
"--          'Location' header that a server sends as part of an\n"
"--          3xx response;\n"
"--\n"
"--      accept_encoding - enables automatic decompression of HTTP\n"
"--          responses;\n"
"--\n"
"--  Returns:\n"
"--      {\n"
"--          status=NUMBER,\n"
"--          reason=ERRMSG\n"
"--          body=STRING,\n"
"--          headers=STRING,\n"
"--          errmsg=STRING\n"
"--      }\n"
"--\n"
"--  Raises error() on invalid arguments and OOM\n"
"--\n"
"\n"
"curl_mt = {\n"
"    __index = {\n"
"        --\n"
"        --  <request> see above <request>\n"
"        --\n"
"        request = function(self, method, url, body, opts)\n"
"            if not method or not url then\n"
"                error('request(method, url[, body, [options]])')\n"
"            end\n"
"            local resp = self.curl:request(method, url, body, opts or {})\n"
"            if resp and resp.headers then\n"
"                if resp.headers['set-cookie'] ~= nil then\n"
"                    resp.cookies = process_cookies(resp.headers['set-cookie'])\n"
"                end\n"
"                resp.headers = process_headers(resp.headers)\n"
"            end\n"
"            return resp\n"
"        end,\n"
"\n"
"        --\n"
"        -- <get> - see <request>\n"
"        --\n"
"        get = function(self, url, options)\n"
"            check_args(self, 'get')\n"
"            return self:request('GET', url, nil, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <post> - see <request>\n"
"        --\n"
"        post = function(self, url, body, options)\n"
"            check_args(self, 'post')\n"
"            return self:request('POST', url, body, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <put> - see <request>\n"
"        --\n"
"        put = function(self, url, body, options)\n"
"            check_args(self, 'put')\n"
"            return self:request('PUT', url, body, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <patch> - see <request>\n"
"        --\n"
"        patch = function(self, url, body, options)\n"
"            check_args(self, 'patch')\n"
"            return self:request('PATCH', url, body, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <options> see <request>\n"
"        --\n"
"        options = function(self, url, options)\n"
"            check_args(self, 'options')\n"
"            return self:request('OPTIONS', url, nil, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <head> see <request>\n"
"        --\n"
"        head = function(self, url, options)\n"
"            check_args(self, 'head')\n"
"            return self:request('HEAD', url, nil, options)\n"
"        end,\n"
"        --\n"
"        -- <delete> see <request>\n"
"        --\n"
"        delete = function(self, url, options)\n"
"            check_args(self, 'delete')\n"
"            return self:request('DELETE', url, nil, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <trace> see <request>\n"
"        --\n"
"        trace = function(self, url, options)\n"
"            check_args(self, 'trace')\n"
"            return self:request('TRACE', url, nil, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <connect> see <request>\n"
"        --\n"
"        connect = function(self, url, options)\n"
"            check_args(self, 'connect')\n"
"            return self:request('CONNECT', url, nil, options)\n"
"        end,\n"
"\n"
"        --\n"
"        -- <stat> - this function returns a table with many values of statistic.\n"
"        --\n"
"        -- Returns {\n"
"        --\n"
"        --  active_requests - this is number of currently executing requests\n"
"        --\n"
"        --  sockets_added -\n"
"        --  this is a total number of added sockets into libev loop\n"
"        --\n"
"        --  sockets_deleted -\n"
"        --  this is a total number of deleted sockets from libev\n"
"        --                      loop\n"
"        --\n"
"        --  total_requests - this is a total number of requests\n"
"        --\n"
"        --  http_200_responses -\n"
"        --  this is a total number of requests which have\n"
"        --              returned a code HTTP 200\n"
"        --\n"
"        --  http_other_responses -\n"
"        --      this is a total number of requests which have\n"
"        --      requests not a HTTP 200\n"
"        --\n"
"        --  failed_requests - this is a total number of requests which have\n"
"        --      failed (included systeme erros, curl errors, HTTP\n"
"        --      errors and so on)\n"
"        --  }\n"
"        --  or error()\n"
"        --\n"
"        stat = function(self)\n"
"            return self.curl:stat()\n"
"        end,\n"
"\n"
"    },\n"
"}\n"
"\n"
"--\n"
"-- Export\n"
"--\n"
"local http_default = http_new()\n"
"local this_module = { new = http_new, }\n"
"\n"
"local function http_default_wrap(fname)\n"
"    return function(...) return http_default[fname](http_default, ...) end\n"
"end\n"
"\n"
"for _, name in ipairs({ 'get', 'delete', 'trace', 'options', 'head',\n"
"                     'connect', 'post', 'put', 'patch', 'request'}) do\n"
"    this_module[name] = http_default_wrap(name)\n"
"end\n"
"\n"
"return this_module\n"
""
;
