const char log_lua[] =
"-- log.lua\n"
"--\n"
"local ffi = require('ffi')\n"
"ffi.cdef[[\n"
"    typedef void (*sayfunc_t)(int level, const char *filename, int line,\n"
"               const char *error, const char *format, ...);\n"
"\n"
"    enum say_logger_type {\n"
"        SAY_LOGGER_BOOT,\n"
"        SAY_LOGGER_STDERR,\n"
"        SAY_LOGGER_FILE,\n"
"        SAY_LOGGER_PIPE,\n"
"        SAY_LOGGER_SYSLOG\n"
"    };\n"
"\n"
"    enum say_logger_type\n"
"    log_type();\n"
"\n"
"    void\n"
"    say_set_log_level(int new_level);\n"
"\n"
"    void\n"
"    say_set_log_format(enum say_format format);\n"
"\n"
"    extern void\n"
"    say_logger_init(const char *init_str, int level, int nonblock,\n"
"                    const char *format, int background);\n"
"\n"
"    extern bool\n"
"    say_logger_initialized(void);\n"
"\n"
"    extern sayfunc_t _say;\n"
"    extern struct ev_loop;\n"
"    extern struct ev_signal;\n"
"\n"
"    extern void\n"
"    say_logrotate(struct ev_loop *, struct ev_signal *, int);\n"
"\n"
"    enum say_level {\n"
"        S_FATAL,\n"
"        S_SYSERROR,\n"
"        S_ERROR,\n"
"        S_CRIT,\n"
"        S_WARN,\n"
"        S_INFO,\n"
"        S_VERBOSE,\n"
"        S_DEBUG\n"
"    };\n"
"\n"
"    enum say_format {\n"
"        SF_PLAIN,\n"
"        SF_JSON\n"
"    };\n"
"    pid_t log_pid;\n"
"    extern int log_level;\n"
"    extern int log_format;\n"
"\n"
"    int\n"
"    say_parse_logger_type(const char **str, enum say_logger_type *type);\n"
"]]\n"
"\n"
"local S_WARN = ffi.C.S_WARN\n"
"local S_INFO = ffi.C.S_INFO\n"
"local S_VERBOSE = ffi.C.S_VERBOSE\n"
"local S_DEBUG = ffi.C.S_DEBUG\n"
"local S_ERROR = ffi.C.S_ERROR\n"
"\n"
"local json = require(\"json\").new()\n"
"json.cfg{\n"
"    encode_invalid_numbers = true,\n"
"    encode_load_metatables = true,\n"
"    encode_use_tostring    = true,\n"
"    encode_invalid_as_nil  = true,\n"
"}\n"
"\n"
"local special_fields = {\n"
"    \"file\",\n"
"    \"level\",\n"
"    \"pid\",\n"
"    \"line\",\n"
"    \"cord_name\",\n"
"    \"fiber_name\",\n"
"    \"fiber_id\",\n"
"    \"error_msg\"\n"
"}\n"
"\n"
"-- Map format number to string.\n"
"local fmt_num2str = {\n"
"    [ffi.C.SF_PLAIN]    = \"plain\",\n"
"    [ffi.C.SF_JSON]     = \"json\",\n"
"}\n"
"\n"
"-- Map format string to number.\n"
"local fmt_str2num = {\n"
"    [\"plain\"]           = ffi.C.SF_PLAIN,\n"
"    [\"json\"]            = ffi.C.SF_JSON,\n"
"}\n"
"\n"
"local function fmt_list()\n"
"    local keyset = {}\n"
"    for k in pairs(fmt_str2num) do\n"
"        keyset[#keyset + 1] = k\n"
"    end\n"
"    return table.concat(keyset, ',')\n"
"end\n"
"\n"
"-- Logging levels symbolic representation.\n"
"local log_level_keys = {\n"
"    ['fatal']       = ffi.C.S_FATAL,\n"
"    ['syserror']    = ffi.C.S_SYSERROR,\n"
"    ['error']       = ffi.C.S_ERROR,\n"
"    ['crit']        = ffi.C.S_CRIT,\n"
"    ['warn']        = ffi.C.S_WARN,\n"
"    ['info']        = ffi.C.S_INFO,\n"
"    ['verbose']     = ffi.C.S_VERBOSE,\n"
"    ['debug']       = ffi.C.S_DEBUG,\n"
"}\n"
"\n"
"local function log_level_list()\n"
"    local keyset = {}\n"
"    for k in pairs(log_level_keys) do\n"
"        keyset[#keyset + 1] = k\n"
"    end\n"
"    return table.concat(keyset, ',')\n"
"end\n"
"\n"
"-- Default options. The keys are part of\n"
"-- user API , so change with caution.\n"
"local log_cfg = {\n"
"    log             = nil,\n"
"    nonblock        = nil,\n"
"    level           = S_INFO,\n"
"    format          = fmt_num2str[ffi.C.SF_PLAIN],\n"
"}\n"
"\n"
"-- Name mapping from box to log module and\n"
"-- back. Make sure all required fields\n"
"-- are covered!\n"
"local log2box_keys = {\n"
"    ['log']             = 'log',\n"
"    ['nonblock']        = 'log_nonblock',\n"
"    ['level']           = 'log_level',\n"
"    ['format']          = 'log_format',\n"
"}\n"
"\n"
"local box2log_keys = {\n"
"    ['log']             = 'log',\n"
"    ['log_nonblock']    = 'nonblock',\n"
"    ['log_level']       = 'level',\n"
"    ['log_format']      = 'format',\n"
"}\n"
"\n"
"-- Update cfg value(s) in box.cfg instance conditionally\n"
"local function box_cfg_update(log_key)\n"
"    -- if it is not yet even exist just exit early\n"
"    if type(box.cfg) ~= 'table' then\n"
"        return\n"
"    end\n"
"\n"
"    local update = function(log_key, box_key)\n"
"        -- the box entry may be under configuration\n"
"        -- process thus equal to nil, skip it then\n"
"        if log_cfg[log_key] ~= nil and\n"
"            box.cfg[box_key] ~= nil and\n"
"            box.cfg[box_key] ~= log_cfg[log_key] then\n"
"            box.cfg[box_key] = log_cfg[log_key]\n"
"        end\n"
"    end\n"
"\n"
"    if log_key == nil then\n"
"        for k, v in pairs(log2box_keys) do\n"
"            update(k, v)\n"
"        end\n"
"    else\n"
"        assert(log2box_keys[log_key] ~= nil)\n"
"        update(log_key, log2box_keys[log_key])\n"
"    end\n"
"end\n"
"\n"
"-- Log options which can be set ony once.\n"
"local cfg_static_keys = {\n"
"    log         = true,\n"
"    nonblock    = true,\n"
"}\n"
"\n"
"-- Test if static key is not changed.\n"
"local function verify_static(k, v)\n"
"    assert(cfg_static_keys[k] ~= nil)\n"
"\n"
"    if ffi.C.say_logger_initialized() == true then\n"
"        if log_cfg[k] ~= v then\n"
"            return false, \"can't be set dynamically\"\n"
"        end\n"
"    end\n"
"\n"
"    return true\n"
"end\n"
"\n"
"local function parse_format(log)\n"
"    -- There is no easy way to get pointer to ponter via FFI\n"
"    local str_p = ffi.new('const char*[1]')\n"
"    str_p[0] = ffi.cast('char*', log)\n"
"    local logger_type = ffi.new('enum say_logger_type[1]')\n"
"    local rc = ffi.C.say_parse_logger_type(str_p, logger_type)\n"
"\n"
"    if rc ~= 0 then\n"
"        box.error()\n"
"    end\n"
"\n"
"    return logger_type[0]\n"
"end\n"
"\n"
"-- Test if format is valid.\n"
"local function verify_format(key, name, cfg)\n"
"    assert(log_cfg[key] ~= nil)\n"
"\n"
"    if not fmt_str2num[name] then\n"
"        local m = \"expected %s\"\n"
"        return false, m:format(fmt_list())\n"
"    end\n"
"\n"
"    local log_type = ffi.C.log_type()\n"
"\n"
"    -- When comes from log.cfg{} or box.cfg{}\n"
"    -- initial call we might be asked to setup\n"
"    -- syslog with json which is not allowed.\n"
"    --\n"
"    -- Note the cfg table comes from two places:\n"
"    -- box api interface and log module itself.\n"
"    -- The good thing that we're only needed log\n"
"    -- entry which is the same key for both.\n"
"    if cfg ~= nil and cfg['log'] ~= nil then\n"
"        log_type = parse_format(cfg['log'])\n"
"    end\n"
"\n"
"    if fmt_str2num[name] == ffi.C.SF_JSON then\n"
"        if log_type == ffi.C.SAY_LOGGER_SYSLOG then\n"
"            local m = \"%s can't be used with syslog logger\"\n"
"            return false, m:format(fmt_num2str[ffi.C.SF_JSON])\n"
"        end\n"
"    end\n"
"\n"
"    return true\n"
"end\n"
"\n"
"-- Test if level is a valid string. The\n"
"-- number may be any for to backward compatibility.\n"
"local function verify_level(key, level)\n"
"    assert(log_cfg[key] ~= nil)\n"
"\n"
"    if type(level) == 'string' then\n"
"        if not log_level_keys[level] then\n"
"            local m = \"expected %s\"\n"
"            return false, m:format(log_level_list())\n"
"        end\n"
"    elseif type(level) ~= 'number' then\n"
"            return false, \"must be a number or a string\"\n"
"    end\n"
"\n"
"    return true\n"
"end\n"
"\n"
"local verify_ops = {\n"
"    ['log']         = verify_static,\n"
"    ['nonblock']    = verify_static,\n"
"    ['format']      = verify_format,\n"
"    ['level']       = verify_level,\n"
"}\n"
"\n"
"-- Verify a value for the particular key.\n"
"local function verify_option(k, v, ...)\n"
"    assert(k ~= nil)\n"
"\n"
"    if verify_ops[k] ~= nil then\n"
"        return verify_ops[k](k, v, ...)\n"
"    end\n"
"\n"
"    return true\n"
"end\n"
"\n"
"-- Main routine which pass data to C logging code.\n"
"local function say(level, fmt, ...)\n"
"    if ffi.C.log_level < level then\n"
"        -- don't waste cycles on debug.getinfo()\n"
"        return\n"
"    end\n"
"    local type_fmt = type(fmt)\n"
"    local format = \"%s\"\n"
"    if select('#', ...) ~= 0 then\n"
"        local stat\n"
"        stat, fmt = pcall(string.format, fmt, ...)\n"
"        if not stat then\n"
"            error(fmt, 3)\n"
"        end\n"
"    elseif type_fmt == 'table' then\n"
"        -- ignore internal keys\n"
"        for _, field in ipairs(special_fields) do\n"
"            fmt[field] = nil\n"
"        end\n"
"        fmt = json.encode(fmt)\n"
"        if ffi.C.log_format == ffi.C.SF_JSON then\n"
"            -- indicate that message is already encoded in JSON\n"
"            format = fmt_num2str[ffi.C.SF_JSON]\n"
"        end\n"
"    elseif type_fmt ~= 'string' then\n"
"        fmt = tostring(fmt)\n"
"    end\n"
"\n"
"    local debug = require('debug')\n"
"    local frame = debug.getinfo(3, \"Sl\")\n"
"    local line, file = 0, 'eval'\n"
"    if type(frame) == 'table' then\n"
"        line = frame.currentline or 0\n"
"        file = frame.short_src or frame.src or 'eval'\n"
"    end\n"
"\n"
"    ffi.C._say(level, file, line, nil, format, fmt)\n"
"end\n"
"\n"
"-- Just a syntactic sugar over say routine.\n"
"local function say_closure(lvl)\n"
"    return function (fmt, ...)\n"
"        say(lvl, fmt, ...)\n"
"    end\n"
"end\n"
"\n"
"-- Rotate log (basically reopen the log file and\n"
"-- start writting into it).\n"
"local function log_rotate()\n"
"    ffi.C.say_logrotate(nil, nil, 0)\n"
"end\n"
"\n"
"-- Set new logging level, the level must be valid!\n"
"local function set_log_level(level, update_box_cfg)\n"
"    assert(type(level) == 'number')\n"
"\n"
"    ffi.C.say_set_log_level(level)\n"
"\n"
"    rawset(log_cfg, 'level', level)\n"
"\n"
"    if update_box_cfg then\n"
"        box_cfg_update('level')\n"
"    end\n"
"\n"
"    local m = \"log: level set to %s\"\n"
"    say(S_DEBUG, m:format(level))\n"
"end\n"
"\n"
"-- Tries to set a new level, or print an error.\n"
"local function log_level(level)\n"
"    local ok, msg = verify_option('level', level)\n"
"    if not ok then\n"
"        error(msg)\n"
"    end\n"
"\n"
"    if type(level) == 'string' then\n"
"        level = log_level_keys[level]\n"
"    end\n"
"\n"
"    set_log_level(level, true)\n"
"end\n"
"\n"
"-- Set a new logging format, the name must be valid!\n"
"local function set_log_format(name, update_box_cfg)\n"
"    assert(fmt_str2num[name] ~= nil)\n"
"\n"
"    if fmt_str2num[name] == ffi.C.SF_JSON then\n"
"        ffi.C.say_set_log_format(ffi.C.SF_JSON)\n"
"    else\n"
"        ffi.C.say_set_log_format(ffi.C.SF_PLAIN)\n"
"    end\n"
"\n"
"    rawset(log_cfg, 'format', name)\n"
"\n"
"    if update_box_cfg then\n"
"        box_cfg_update('format')\n"
"    end\n"
"\n"
"    local m = \"log: format set to '%s'\"\n"
"    say(S_DEBUG, m:format(name))\n"
"end\n"
"\n"
"-- Tries to set a new format, or print an error.\n"
"local function log_format(name)\n"
"    local ok, msg = verify_option('format', name)\n"
"    if not ok then\n"
"        error(msg)\n"
"    end\n"
"\n"
"    set_log_format(name, true)\n"
"end\n"
"\n"
"-- Returns pid of a pipe process.\n"
"local function log_pid()\n"
"    return tonumber(ffi.C.log_pid)\n"
"end\n"
"\n"
"-- Fetch a value from log to box.cfg{}.\n"
"local function box_api_cfg_get(key)\n"
"    return log_cfg[box2log_keys[key]]\n"
"end\n"
"\n"
"-- Set value to log from box.cfg{}.\n"
"local function box_api_cfg_set(cfg, key, value)\n"
"    local log_key = box2log_keys[key]\n"
"\n"
"    -- a special case where we need to restore\n"
"    -- nil value from previous setup attempt.\n"
"    if value == box.NULL then\n"
"        log_cfg[log_key] = nil\n"
"        return true\n"
"    end\n"
"\n"
"    local ok, msg = verify_option(log_key, value, cfg)\n"
"    if not ok then\n"
"        return false, msg\n"
"    end\n"
"\n"
"    log_cfg[log_key] = value\n"
"    return true\n"
"end\n"
"\n"
"-- Set logging level from reloading box.cfg{}\n"
"local function box_api_cfg_set_log_level()\n"
"    local log_key = box2log_keys['log_level']\n"
"    local v = box.cfg['log_level']\n"
"\n"
"    local ok, msg = verify_option(log_key, v)\n"
"    if not ok then\n"
"        return false, msg\n"
"    end\n"
"\n"
"    if type(v) == 'string' then\n"
"        v = log_level_keys[v]\n"
"    end\n"
"\n"
"    set_log_level(v, false)\n"
"    return true\n"
"end\n"
"\n"
"-- Set logging format from reloading box.cfg{}\n"
"local function box_api_set_log_format()\n"
"    local log_key = box2log_keys['log_format']\n"
"    local v = box.cfg['log_format']\n"
"\n"
"    local ok, msg = verify_option(log_key, v)\n"
"    if not ok then\n"
"        return false, msg\n"
"    end\n"
"\n"
"    set_log_format(v, false)\n"
"    return true\n"
"end\n"
"\n"
"-- Reload dynamic options.\n"
"local function reload_cfg(cfg)\n"
"    for k in pairs(cfg_static_keys) do\n"
"        if cfg[k] ~= nil then\n"
"            local ok, msg = verify_static(k, cfg[k])\n"
"            if not ok then\n"
"                local m = \"log.cfg: \\'%s\\' %s\"\n"
"                error(m:format(k, msg))\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    if cfg.level ~= nil then\n"
"        log_level(cfg.level)\n"
"    end\n"
"\n"
"    if cfg.format ~= nil then\n"
"        log_format(cfg.format)\n"
"    end\n"
"end\n"
"\n"
"-- Load or reload configuration via log.cfg({}) call.\n"
"local function load_cfg(self, cfg)\n"
"    cfg = cfg or {}\n"
"\n"
"    -- log option might be zero length string, which\n"
"    -- is fine, we should treat it as nil.\n"
"    if cfg.log ~= nil then\n"
"        if type(cfg.log) ~= 'string' or cfg.log:len() == 0 then\n"
"            cfg.log = nil\n"
"        end\n"
"    end\n"
"\n"
"    if cfg.format ~= nil then\n"
"        local ok, msg = verify_option('format', cfg.format, cfg)\n"
"        if not ok then\n"
"            local m = \"log.cfg: \\'%s\\' %s\"\n"
"            error(m:format('format', msg))\n"
"        end\n"
"    end\n"
"\n"
"    if cfg.level ~= nil then\n"
"        local ok, msg = verify_option('level', cfg.level)\n"
"        if not ok then\n"
"            local m = \"log.cfg: \\'%s\\' %s\"\n"
"            error(m:format('level', msg))\n"
"        end\n"
"    end\n"
"\n"
"    if cfg.nonblock ~= nil then\n"
"        if type(cfg.nonblock) ~= 'boolean' then\n"
"            error(\"log.cfg: 'nonblock' option must be 'true' or 'false'\")\n"
"        end\n"
"    end\n"
"\n"
"    if ffi.C.say_logger_initialized() == true then\n"
"        return reload_cfg(cfg)\n"
"    end\n"
"\n"
"    cfg.level = cfg.level or log_cfg.level\n"
"    cfg.format = cfg.format or log_cfg.format\n"
"    cfg.nonblock = cfg.nonblock or log_cfg.nonblock\n"
"\n"
"    -- nonblock is special: it has to become integer\n"
"    -- for ffi call but in config we have to save\n"
"    -- true value only for backward compatibility!\n"
"    local nonblock = cfg.nonblock\n"
"\n"
"    if nonblock == nil or nonblock == false then\n"
"        nonblock = 0\n"
"    else\n"
"        nonblock = 1\n"
"    end\n"
"\n"
"    -- Parsing for validation purposes\n"
"    if cfg.log ~= nil then\n"
"        parse_format(cfg.log)\n"
"    end\n"
"\n"
"    -- We never allow confgure the logger in background\n"
"    -- mode since we don't know how the box will be configured\n"
"    -- later.\n"
"    ffi.C.say_logger_init(cfg.log, cfg.level,\n"
"                          nonblock, cfg.format, 0)\n"
"\n"
"    if nonblock == 1 then\n"
"        nonblock = true\n"
"    else\n"
"        nonblock = nil\n"
"    end\n"
"\n"
"    -- Update log_cfg vars to show them in module\n"
"    -- configuration output.\n"
"    rawset(log_cfg, 'log', cfg.log)\n"
"    rawset(log_cfg, 'level', cfg.level)\n"
"    rawset(log_cfg, 'nonblock', nonblock)\n"
"    rawset(log_cfg, 'format', cfg.format)\n"
"\n"
"    -- and box.cfg output as well.\n"
"    box_cfg_update()\n"
"\n"
"    local m = \"log.cfg({log=%s,level=%s,nonblock=%s,format=\\'%s\\'})\"\n"
"    say(S_DEBUG, m:format(cfg.log, cfg.level, cfg.nonblock, cfg.format))\n"
"end\n"
"\n"
"local compat_warning_said = false\n"
"local compat_v16 = {\n"
"    logger_pid = function()\n"
"        if not compat_warning_said then\n"
"            compat_warning_said = true\n"
"            say(S_WARN, 'logger_pid() is deprecated, please use pid() instead')\n"
"        end\n"
"        return log_pid()\n"
"    end;\n"
"}\n"
"\n"
"local log = {\n"
"    warn = say_closure(S_WARN),\n"
"    info = say_closure(S_INFO),\n"
"    verbose = say_closure(S_VERBOSE),\n"
"    debug = say_closure(S_DEBUG),\n"
"    error = say_closure(S_ERROR),\n"
"    rotate = log_rotate,\n"
"    pid = log_pid,\n"
"    level = log_level,\n"
"    log_format = log_format,\n"
"    cfg = setmetatable(log_cfg, {\n"
"        __call = load_cfg,\n"
"    }),\n"
"    box_api = {\n"
"        cfg_get = box_api_cfg_get,\n"
"        cfg_set = box_api_cfg_set,\n"
"        cfg_set_log_level = box_api_cfg_set_log_level,\n"
"        cfg_set_log_format = box_api_set_log_format,\n"
"    },\n"
"}\n"
"\n"
"setmetatable(log, {\n"
"    __serialize = function(self)\n"
"        local res = table.copy(self)\n"
"        res.box_api = nil\n"
"        return setmetatable(res, {})\n"
"    end,\n"
"    __index = compat_v16;\n"
"})\n"
"\n"
"return log\n"
""
;
